''***************************************
''*	PS/2 Keyboard Driver v1.0.1	*
''*	Author: Chip Gracey		*
''*	Copyright (c) 2004 Parallax, Inc.	*
''*	See end of file for terms of use.	*
''***************************************

{-----------------REVISION HISTORY-----------------
 v1.0.1 - Updated 6/15/2006 to work with Propeller Tool 0.96}

CON
	stat_extended	= 1		' extended keycode flag
	stat_break	= 2		' break code flag
	stat_config	= 4		' configure keyboard flag
	stat_reset	= 8		' reset keyboard flag

VAR
	long	cog
	long	par_states[8]	'key states (256 bits)	read-only
	long	par_present	'keyboard present	read-only

PUB start(dpin, cpin) : okay

'' Start keyboard driver - starts a cog
'' returns false if no cog available
''
''	dpin	= data signal on PS/2 jack
''	cpin	= clock signal on PS/2 jack
''
''	use 100-ohm resistors between pins and jack
''	use 10K-ohm resistors to pull jack-side signals to VDD
''	connect jack-power to 5V, jack-gnd to VSS
''
'' all lock-keys will be enabled, NumLock will be initially 'on',
'' and auto-repeat will be set to 15cps with a delay of .5s

	okay := startx(dpin, cpin, %0_000_100, %01_01000, 0, 0)

PUB startx(dpin, cpin, locks, auto, irq, rtc) : okay

'' Like start, but allows you to specify lock settings and auto-repeat
''
''	locks = lock setup
''		bit 6 disallows shift-alphas (case set soley by CapsLock)
''		bits 5..3 disallow toggle of NumLock/CapsLock/ScrollLock state
''		bits 2..0 specify initial state of NumLock/CapsLock/ScrollLock
''		(eg. %0_001_100 = disallow ScrollLock, NumLock initially 'on')
''
''	auto	= auto-repeat setup
''		bits 6..5 specify delay (0=.25s, 1=.5s, 2=.75s, 3=1s)
''		bits 4..0 specify repeat rate (0=30cps..31=2cps)
''		(eg %01_00000 = .5s delay, 30cps repeat)

	stop
	longmove(@par_states, @dpin, 6)
	okay := cog := cognew(@entry, @par_states) + 1


PUB stop
'' Stop keyboard driver - frees a cog
	if cog
		cogstop(cog~ -	1)
	longfill(@par_states, 0, 8+1+1)


PUB present : truefalse
'' Check if keyboard present - valid ~2s after start
'' returns t|f
	truefalse := -par_present

PUB keystate(k) : state
'' Get the state of a particular key
'' returns t|f
	state := -(par_states[k >> 5] >> k & 1)

PUB states_ptr
	return @par_states	' returns the location of par_states longs

DAT

'******************************************
'* Assembly language PS/2 keyboard driver *
'******************************************

		org	0
'
'
' Entry
'
entry		movd	:par, #_dpin			' load input parameters _dpin/_cpin/_locks/_auto/_irq/_rtc
		mov	x, par
		mov	y, #6
		mov	t, #0
:par		rdlong	0, x
		add	:par, dlsb
		wrlong	t, x				' clear the states
		add	x, #4
		djnz	y, #:par

		mov	dmask, #1			' set pin masks
		shl	dmask, _dpin			' 0..31 (ignores bit #5)
		mov	cmask, #1
		shl	cmask, _cpin			' 0..31 (ignores bit #5)

		rdlong	clkfreq256, #0			' get clkfreq
		shr	clkfreq256, #8			' /256

		mov	cnt0, CNT
		mov	irqctr, #0

		test	_dpin, #$20		WC	' modify port registers within code
		muxc	_d1, dlsb
		muxc	_d2, dlsb
		muxc	_d3, #1
		muxc	_d4, #1
		test	_cpin, #$20		WC
		muxc	_c1, dlsb
		muxc	_c2, dlsb
		muxc	_c3, #1
'
'
' Reset keyboard
'
reset		mov	dira, #0			' reset directions
		mov	dirb, #0

		movd	:par, #_states			' reset output parameters _states[8]/_present
		mov	x, #8+1
:par		mov	0, #0
		add	:par, dlsb
		djnz	x, #:par

		mov	stat, #stat_reset		' set reset flag
'
'
' Update parameters
'
update		movd	:par, #_states			' update output _states[8]/_present
		mov	x, par
		mov	y, #8+1
:par		wrlong	0, x
		add	:par, dlsb
		add	x, #4
		djnz	y, #:par

		test	stat,#stat_reset	WC	' if reset flag, transmit reset command
	if_c	mov	data,#$ff
	if_c	call	#transmit
'
'
' Get scancode
'
newcode
		mov	stat, #0			' reset state

:same
		call	#receive			' receive byte from keyboard

		cmp	data, #$83+1		WC	' scancode?

 if_nc		cmp	data, #$AA		WZ	' powerup/reset?
 if_nc_and_z	jmp	#configure

 if_nc		cmp	data, #$e0		WZ	' extended?
 if_nc_and_z	or	stat, #stat_extended
 if_nc_and_z	jmp	#:same

 if_nc		cmp	data, #$f0		WZ	' released?
 if_nc_and_z	or	stat, #stat_break
 if_nc_and_z	jmp	#:same

 if_nc		jmp	#newcode			' unknown, ignore
'
'
' Translate scancode and enter into buffer
'
		test	stat, #stat_extended	WC	' lookup code with extended flag
		rcl	data, #1
		call	#look

		cmp	data, #$ff		WZ	' if unknown, ignore
	if_z	jmp	#newcode

		mov	t,_states+6			' remember lock keys in _states

		mov	x, data				' set/clear key bit in _states
		shr	x, #5
		add	x, #_states
		movd	:reg, x
		mov	y, #1
		shl	y, data
		test	stat,#stat_break	WC
:reg		muxnc	0,y

		test	stat, #stat_config	WC	' if not configure flag, done
	if_nc	jmp	#update				' else configure to update leds
'
'
' Configure keyboard
'
configure
		mov	data,#$F3			' set keyboard auto-repeat
		call	#transmit
		mov	data,_auto
		and	data,#%11_11111
		call	#transmit

		mov	data,#$ED			' set keyboard lock-leds
		call	#transmit
		mov	data,_locks
		rev	data,#-3 & $1F
		test	data,#%100		WC
		rcl	data,#1
		and	data,#%111
		call	#transmit

		mov	x,_locks			' insert locks into _states
		and	x,#%111
		shl	_states+7,#3
		or	_states+7,x
		ror	_states+7,#3

		mov	_present,#1			' set _present

		jmp	#update				' done
'
'
' Lookup byte in table
'
look
		ror	data, #2			' perform lookup
		movs	:reg, data
		add	:reg, #table
		shr	data, #27
		mov	x, data
:reg		mov	data, 0-0
		shr	data, x
		jmp	#rand				' isolate byte
'
'
' Transmit byte to keyboard
'
transmit
_c1		or	dira,cmask			' pull clock low
		movs	napshr,#13			' hold clock for ~128us (must be >100us)
		call	#nap
_d1		or	dira,dmask			' pull data low
		movs	napshr,#18			' hold data for ~4us
		call	#nap
_c2		xor	dira,cmask			' release clock

		test	data,#$0FF		WC	' append parity and stop bits to byte
		muxnc	data,#$100
		or	data,dlsb

		mov	x,#10				' ready 10 bits
transmit_bit
		call	#wait_c0			' wait until clock low
		shr	data,#1			WC	' output data bit
_d2		muxnc	dira,dmask
		mov	wcond,c1			' wait until clock high
		call	#wait
		djnz	x, #transmit_bit		' another bit?

		mov	wcond,c0d0			' wait until clock and data low
		call	#wait
		mov	wcond,c1d1			' wait until clock and data high
		call	#wait

		call	#receive_ack			' receive ack byte with timed wait
		cmp	data,#$FA		WZ	' if ack error, reset keyboard
	if_nz	jmp	#reset

transmit_ret
		ret
'
'
' Receive byte from keyboard
'
receive		test	_cpin, #$20		WC	' wait indefinitely for initial clock low
:wait		test	cmask, ina		WZ
	if_z	jmp	#receive_ack
		mov	ctr, CNT
		sub	ctr, cnt0
		cmps	ctr, clkfreq256 	WZ, WC	' >= 1/256 second?
	if_ae	call	#do_irq				' generate an interrupt request, count RTC
		jmp	#:wait
receive_ack
		mov	x,#11				' ready 11 bits
receive_bit
		call	#wait_c0			' wait until clock low
		movs	napshr, #16			' pause ~16us
		call	#nap

_d3		test	dmask, ina		WC	' input data bit
		rcr	data, #1
		mov	wcond, c1			' wait until clock high
		call	#wait
		djnz	x, #receive_bit			' another bit?

		shr	data, #22			' align byte
		test	data, #$1FF		WC	' if parity error, reset keyboard
	if_nc	jmp	#reset
rand		and	data, #$FF			' isolate byte

look_ret
receive_ack_ret
receive_ret
		ret
'
'
' Wait for clock/data to be in required state(s)
'
wait_c0		mov	wcond, c0			' (wait until clock low)

wait		mov	y, tenms			' set timeout to 10ms

wloop		movs	napshr, #18			' nap ~4us
		call	#nap
_c3		test	cmask, ina		WC	' check required state(s)
_d4		test	dmask, ina		WZ	' loop until got state(s) or timeout
wcond
 if_never	djnz	y, #wloop			' (replaced with c0/c1/c0d0/c1d1)
		tjz	y, #reset			' if timeout, reset keyboard
wait_ret
wait_c0_ret
		ret


c0	if_c		djnz	y, #wloop		' (if_never replacements)
c1	if_nc		djnz	y, #wloop
c0d0	if_c_or_nz	djnz	y, #wloop
c1d1	if_nc_or_z	djnz	y, #wloop
'
'
' Nap
'
nap		rdlong	t, #0				' get clkfreq
napshr		shr	t, #18/16/13			' shr scales time
		min	t, #3				' ensure waitcnt won't snag
		add	t, cnt				' add cnt to time
		waitcnt t, #0				' wait until time elapses (nap)

nap_ret
		ret

'
' Generate interrupt request and once per second update the RTC
'
do_irq
		add	irqctr, #1
		add	cnt0, clkfreq256		' next 1/256 second
		tjz	_irq, #do_irq_ret		' _irq not set, so just return

		rdlong	ireq, _irq			' get irq
		or	ireq, #%01000000		' request RST 7.5
		test	irqctr, #255	WZ		' one second?
		wrlong	ireq, _irq
	if_nz	jmp	#do_irq_ret			' not yet

		cmp	_rtc, #0	WZ		' is _rtc set to a memory address?
	if_z	jmp	#do_irq_ret			' no, just return

		mov	ptr, _rtc			' pointer to RTC bytes
		rdbyte	bcd, ptr			' get RTC seconds
		call	#bcd_inc
		cmp	bcd, #$60	WZ, WC
	if_ae	mov	bcd, #$00			' seconds wrap
		wrbyte	bcd, ptr
	if_b	jmp	#do_irq_ret

		add	ptr, #1				' point to minutes
		rdbyte	bcd, ptr
		call	#bcd_inc
		cmp	bcd, #$60	WZ, WC
	if_ae	mov	bcd, #$00			' minutes wrap
		wrbyte	bcd, ptr
	if_b	jmp	#do_irq_ret
	
		add	ptr, #1				' point to hours
		rdbyte	bcd, ptr
		call	#bcd_inc
		cmp	bcd, #$24	WZ, WC
	if_ae	mov	bcd, #$00			' hours wrap
		wrbyte	bcd, ptr
	if_b	jmp	#do_irq_ret
	
		add	ptr, #1				' point to day of month
		rdbyte	bcd, ptr
		call	#bcd_inc
		add	ptr, #1				' point to day of week and month
		rdbyte	mlen, ptr
		shr	mlen, #4			' mask month
		add	mlen, mlength_ptr
		rdbyte	mlen, mlen
		cmp	bcd, mlen	WZ, WC
	if_ae	mov	bcd, #$01			' day of month wrap
		sub	ptr, #1				' point to day of month
		wrbyte	bcd, ptr
	if_b	jmp	#do_irq_ret
	
		add	ptr, #1				' point to day of week and month
		rdbyte	bcd, ptr
		mov	low, bcd
		and	low, #$0f
		shr	bcd, #4
		add	low, #1				' new day of week
	if_ae	add	bcd, #1				' new month
		cmp	low, #7		WZ, WC
	if_ae	mov	low, #$00			' wrap day of week
		cmp	bcd, #12+1	WZ, WC
	if_ae	mov	bcd, #$01			' wrap month
		shl	bcd, #4
		or	bcd, low
		wrbyte	bcd, ptr
do_irq_ret
		ret

bcd_inc
		add	bcd, #1
		mov	low, bcd
		and	low, #%1111
		cmp	low, #10	WZ, WC
	if_ae	add	bcd, #$06
		cmp	bcd, #$99+1	WZ, WC
	if_ae	add	bcd, #$60
bcd_inc_ret
		ret

'
'
' Initialized data
'
'
dlsb			long	1 << 9
tenms			long	10_000 / 4
mlength_ptr		long	@@@mlength
'
'***************************************************************************
'*   w/o SHIFT                             with SHIFT
'*   +-------------------------------+     +-------------------------------+
'*   | 0   1   2   3   4   5   6   7 |     | 0   1   2   3   4   5   6   7 |
'*+--+---+---+---+---+---+---+---+---+  +--+---+---+---+---+---+---+---+---+
'*|0 | z | x | c | v | b | n | m | l |  |0 | Z | X | C | V | B | N | M | L |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|1 | a | s | d | f | g | h | j | k |  |1 | A | S | D | F | G | H | J | K |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|2 | q | w | e | r | t | y | u | i |  |2 | Q | W | E | R | T | Y | U | I |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|3 | o | p | [ | ; | ' | , | . | / |  |3 | O | P | ] | : | " | < | > | ? |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|4 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 |  |4 | ! | @ | # | $ | % | ^ | & | * |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|5 | 9 | 0 | - | = |LFT|RGT|UP |DN |  |5 | ( | ) | _ | + |LFT|RGT|UP |DN |
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|6 |SPC|BSP|TAB|ESC|PST|LBL|PRT|ENT|  |6 |SPC|BSP|TAB|ESC|PST|LBL|PRT|ENT|
'*|  +---+---+---+---+---+---+---+---+  |  +---+---+---+---+---+---+---+---+
'*|7 |F1 |F2 |F3 |F4 |F5 |F6 |F7 |F8 |  |7 |F1 |F2 |F3 |F4 |F5 |F6 |F7 |F8 |
'*+  +---+---+---+---+---+---+---+---+  +  +---+---+---+---+---+---+---+---+
'*|8 |SHF|CTL|GRP|COD|---|CPS|---|PAU|  |8 |SHF|CTL|GRP|COD|---|CPS|---|PAU|
'*+--+---+---+---+---+---+---+---+---+  +--+---+---+---+---+---+---+---+---+
'***************************************************************************/
'
'
' Lookup table
'				reg  ext	scan	extkey	regkey	()=keypad
'
table			byte	$ff, $ff	'00
			byte	$3c, $ff	'01		F9
			byte	$ff, $ff	'02
			byte	$3c, $ff	'03		F5
			byte	$3a, $ff	'04		F3
			byte	$38, $ff	'05		F1
			byte	$39, $ff	'06		F2
			byte	$47, $ff	'07		F12
			byte	$ff, $ff	'08
			byte	$3d, $ff	'09		F10
			byte	$3f, $ff	'0A		F8
			byte	$3d, $ff	'0B		F6
			byte	$3b, $ff	'0C		F4
			byte	$32, $ff	'0D		Tab
			byte	$24, $ff	'0E		`
			byte	$ff, $ff	'0F
			byte	$ff, $ff	'10
			byte	$42, $43	'11	Alt-R	Alt-L
			byte	$40, $ff	'12		Shift-L
			byte	$ff, $ff	'13
			byte	$41, $41	'14	Ctrl-R	Ctrl-L
			byte	$10, $ff	'15		q
			byte	$20, $ff	'16		1
			byte	$ff, $ff	'17
			byte	$ff, $ff	'18
			byte	$ff, $ff	'19
			byte	$00, $ff	'1A		z
			byte	$09, $ff	'1B		s
			byte	$08, $ff	'1C		a
			byte	$11, $ff	'1D		w
			byte	$21, $ff	'1E		2
			byte	$ff, $43	'1F	Win-L
			byte	$ff, $ff	'20
			byte	$02, $ff	'21		c
			byte	$01, $ff	'22		x
			byte	$0a, $ff	'23		d
			byte	$12, $ff	'24		e
			byte	$23, $ff	'25		4
			byte	$22, $ff	'26		3
			byte	$ff, $43	'27	Win-R
			byte	$ff, $ff	'28
			byte	$30, $ff	'29		Space
			byte	$03, $ff	'2A		v
			byte	$0b, $ff	'2B		f
			byte	$14, $ff	'2C		t
			byte	$13, $ff	'2D		r
			byte	$24, $ff	'2E		5
			byte	$ff, $44	'2F	Apps
			byte	$ff, $ff	'30
			byte	$05, $ff	'31		n
			byte	$04, $ff	'32		b
			byte	$0d, $ff	'33		h
			byte	$0c, $ff	'34		g
			byte	$15, $ff	'35		y
			byte	$25, $ff	'36		6
			byte	$ff, $46	'37	Power
			byte	$ff, $ff	'38
			byte	$ff, $ff	'39
			byte	$06, $ff	'3A		m
			byte	$0e, $ff	'3B		j
			byte	$16, $ff	'3C		u
			byte	$26, $ff	'3D		7
			byte	$27, $ff	'3E		8
			byte	$ff, $47	'3F	Sleep
			byte	$ff, $ff	'40
			byte	$1d, $ff	'41		,
			byte	$0f, $ff	'42		k
			byte	$17, $ff	'43		i
			byte	$18, $ff	'44		o
			byte	$29, $ff	'45		0
			byte	$28, $ff	'46		9
			byte	$ff, $ff	'47
			byte	$ff, $ff	'48
			byte	$1e, $ff	'49		.
			byte	$1f, $1f	'4A	(/)	/
			byte	$07, $ff	'4B		l
			byte	$1b, $ff	'4C		;
			byte	$19, $ff	'4D		p
			byte	$2a, $ff	'4E		-
			byte	$ff, $ff	'4F
			byte	$ff, $ff	'50
			byte	$ff, $ff	'51
			byte	$1c, $ff	'52		'
			byte	$ff, $ff	'53
			byte	$1a, $ff	'54		[
			byte	$2b, $ff	'55		=
			byte	$ff, $ff	'56
			byte	$ff, $ff	'57
			byte	$45, $ff	'58		CapsLock
			byte	$40, $ff	'59		Shift-R
			byte	$37, $37	'5A	(Enter) Enter
			byte	$1d, $ff	'5B		]
			byte	$ff, $ff	'5C
			byte	$48, $ff	'5D		\
			byte	$ff, $47	'5E	WakeUp
			byte	$ff, $ff	'5F
			byte	$ff, $ff	'60
			byte	$ff, $ff	'61
			byte	$ff, $ff	'62
			byte	$ff, $ff	'63
			byte	$ff, $ff	'64
			byte	$ff, $ff	'65
			byte	$31, $ff	'66		BackSpace
			byte	$ff, $ff	'67
			byte	$ff, $ff	'68
			byte	$20, $ff	'69	End	(1)
			byte	$ff, $ff	'6A
			byte	$23, $2c	'6B	Left	(4)
			byte	$26, $ff	'6C	Home	(7)
			byte	$ff, $ff	'6D
			byte	$ff, $ff	'6E
			byte	$ff, $ff	'6F
			byte	$29, $ff	'70	Insert	(0)
			byte	$1e, $ff	'71	Delete	(.)
			byte	$21, $2f	'72	Down	(2)
			byte	$24, $ff	'73		(5)
			byte	$25, $2d	'74	Right	(6)
			byte	$27, $2e	'75	Up	(8)
			byte	$33, $ff	'76		Esc
			byte	$ff, $ff	'77		NumLock
			byte	$3e, $ff	'78		F11
			byte	$2b, $ff	'79		(+)
			byte	$22, $ff	'7A	PageDn	(3)
			byte	$2a, $ff	'7B		(-)
			byte	$27, $ff	'7C	PrScr	(*)
			byte	$28, $ff	'7D	PageUp	(9)
			byte	$ff, $ff	'7E		ScrLock
			byte	$ff, $ff	'7F
			byte	$ff, $ff	'80
			byte	$ff, $ff	'81
			byte	$ff, $ff	'82
			byte	$3e, $ff	'83		F7
'
'
' Uninitialized data
'
dmask			res	1
cmask			res	1
stat			res	1
data			res	1
x			res	1
y			res	1
t			res	1
clkfreq256		res	1	' clkfreq / 256
cnt0			res	1	' next CNT value to wait for
irqctr			res	1	' interrupt counter value
ctr			res	1	' counter value
ptr			res	1	' pointer to RTC bytes
ireq			res	1	' interrupt request long
bcd			res	1	' BCD value
low			res	1	' LSB of BCD value
mlen			res	1

_states			res	8	' write-only
_present		res	1	' write-only
_dpin			res	1	' read-only at start
_cpin			res	1	' read-only at start
_locks			res	1	' read-only at start
_auto			res	1	' read-only at start
_irq			res	1	' read-only at start
_rtc			res	1	' read-only at start

			fit	$1f0
mlength			byte	0,$31+1,$28+1,$31+1,$30+1,$31+1,$30+1,$31+1,$31+1,$30+1,$31+1,$30+1,$31+1,0,0,0
{{

+------------------------------------------------------------------------------------------------------------------------------+
|                                                   TERMS OF USE: MIT License                                                  | +------------------------------------------------------------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation    | |files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,    |
|modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software|
|is furnished to do so, subject to the following conditions:                                                                   |
|                                                                                                                              |
|The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.|
|                                                                                                                              |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE          |
|WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR         |
|COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,   |
|ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.                         |
+------------------------------------------------------------------------------------------------------------------------------+
}}
